# pylint: disable=redefined-outer-name,protected-access
from collections import OrderedDict
import string
import random
import os.path

import boto
import pytest

from z3.config import get_config
from z3.snap import (list_snapshots, S3SnapshotManager, ZFSSnapshotManager,
                     PairManager, CommandExecutor, IntegrityError)


class FakeKey(object):
    def __init__(self, name, metadata=None):
        self.name = name
        self.key = name
        self.metadata = metadata


class FakeBucket(object):
    rand_prefix = 'test-' + ''.join([random.choice(string.ascii_letters) for _ in xrange(8)]) + '/'
    fake_data = {
        "pool/fs@snap_0": {'parent': 'pool/fs@snap_expired'},
        "pool/fs@snap_1_f": {'is_full': 'true'},
        "pool/fs@snap_2": {'parent': 'pool/fs@snap_1_f'},
        "pool/fs@snap_3": {'parent': 'pool/fs@snap_2', 'is_full': 'false'},
        "pool/fs@snap_4_mp": {'parent': 'missing_parent'},  # missing parent
        "pool/fs@snap_5": {'parent': 'pool/fs@snap_4_mp'},
        "pool/fs@snap_6_cycle": {'parent': 'pool/fs@snap_7_cycle'},  # cycle
        "pool/fs@snap_7_cycle": {'parent': 'pool/fs@snap_6_cycle'},  # cycle
    }

    def list(self, *a, **kwa):
        # boto bucket.list gives you keys without metadata, let's emulate that
        return (FakeKey(os.path.join(self.rand_prefix, name)) for name in self.fake_data.iterkeys())

    def get_key(self, key):
        name = key[len(self.rand_prefix):]
        return FakeKey(
            name=key,
            metadata=self.fake_data[name])


def write_s3_data():
    """Takes the default data from FakeBucket and writes it to S3.
    Allows running the same tests against fakes and the boto api.
    """
    cfg = get_config()
    bucket = boto.connect_s3(
        cfg['S3_KEY_ID'], cfg['S3_SECRET']).get_bucket(cfg['BUCKET'])
    for name, metadata in FakeBucket.fake_data.iteritems():
        key = bucket.new_key(os.path.join(FakeBucket.rand_prefix, name))
        headers = {("x-amz-meta-" + k): v for k, v in metadata.iteritems()}
        key.set_contents_from_string("spam", headers=headers)
    return bucket


@pytest.fixture(
    scope='module',
    params=[
        FakeBucket,
        pytest.mark.with_s3(write_s3_data),
    ],
    ids=['fake_bucket', 'with_s3']
)
def s3_manager(request):
    """This parametrized fixture will cause any test using it to execute twice,
    once using fakes and again using boto and hitting s3.

    The tests don't actually write to the bucket so we can share the same S3SnapshotManager
    across all tests.
    """
    return S3SnapshotManager(
        request.param(), s3_prefix=FakeBucket.rand_prefix, snapshot_prefix="pool/fs@snap_")


def test_list_snapshots(s3_manager):
    snapshots = sorted(s3_manager.list(), key=lambda el: el.name)
    expected = sorted(
        ["pool/fs@snap_0", "pool/fs@snap_1_f", "pool/fs@snap_2", "pool/fs@snap_3",
         "pool/fs@snap_4_mp", "pool/fs@snap_5", "pool/fs@snap_6_cycle",
         "pool/fs@snap_7_cycle"])
    assert [s.name for s in snapshots] == expected


def test_healthy_full(s3_manager):
    snap = s3_manager.get('pool/fs@snap_1_f')
    assert snap.is_full
    assert snap.is_healthy


def test_healthy_incremental(s3_manager):
    snap = s3_manager.get('pool/fs@snap_3')
    assert snap.is_full is False
    assert snap.is_healthy


def test_unhealthy_incremental(s3_manager):
    snap = s3_manager.get('pool/fs@snap_5')
    assert snap.is_full is False
    assert snap.is_healthy is False
    assert snap.reason_broken == 'parent broken'
    assert snap.parent.reason_broken == 'missing parent'


def test_unhealthy_cycle(s3_manager):
    snap = s3_manager.get('pool/fs@snap_7_cycle')
    assert snap.is_full is False
    assert snap.is_healthy is False
    assert snap.reason_broken == 'cycle detected'
    assert snap.parent.reason_broken == 'cycle detected'


class FakeZFSManager(ZFSSnapshotManager):
    _expected = (
        # pool is a different zfs dataset, the s3 fixtures don't include it
        'pool@snap_p1\t0\t19K\t-\t19K\n'
        'pool@snap_p2\t0\t19K\t-\t0\n'

        'pool/fs@snap_0\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_1_f\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@funky_name\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_2\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_3\t10.0M\t10.0M\t-\t10.0M\n'

        # local only snapshots, not in s3 fixtures
        'pool/fs@snap_8\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_9\t10.0M\t10.0M\t-\t10.0M\n'
    )

    def __init__(self, expected=None, *a, **kwa):
        if expected is not None:
            self._expected = expected
        super(FakeZFSManager, self).__init__(*a, **kwa)

    def _list_snapshots(self):
        return self._expected


def test_list_local_snapshots():
    zfs = FakeZFSManager(fs_name='pool/fs', snapshot_prefix='snap_')
    expected = {
        'pool': OrderedDict([  # _parse_snapshots returns snapshots for ALL filesystems
            ('snap_p1', {
                'name': 'pool@snap_p1',
                'mountpoint': '-', 'refer': '19K', 'used': '0', 'written': '19K',
            }),
            ('snap_p2', {
                'name': 'pool@snap_p2',
                'mountpoint': '-', 'refer': '19K', 'used': '0', 'written': '0',
            }),
        ]),
        'pool/fs': OrderedDict([
            ('snap_0', {
                'name': 'pool/fs@snap_0',
                'mountpoint': '-', 'refer': '10.0M', 'used': '10.0M', 'written': '10.0M',
            }),
            ('snap_1_f', {
                'name': 'pool/fs@snap_1_f',
                'mountpoint': '-', 'refer': '10.0M', 'used': '10.0M', 'written': '10.0M',
            }),
            ('funky_name', {
                'name': 'pool/fs@funky_name',
                'mountpoint': '-', 'refer': '10.0M', 'used': '10.0M', 'written': '10.0M',
            }),
            ('snap_2', {
                'name': 'pool/fs@snap_2',
                'mountpoint': '-', 'refer': '10.0M', 'used': '10.0M', 'written': '10.0M',
            }),
            ('snap_3', {
                'name': 'pool/fs@snap_3',
                'mountpoint': '-', 'refer': '10.0M', 'used': '10.0M', 'written': '10.0M'
            }),
            ('snap_8', {
                'name': 'pool/fs@snap_8',
                'mountpoint': '-', 'refer': '10.0M', 'used': '10.0M', 'written': '10.0M'
            }),
            ('snap_9', {
                'name': 'pool/fs@snap_9',
                'mountpoint': '-', 'refer': '10.0M', 'used': '10.0M', 'written': '10.0M'
            }),
        ])
    }
    snapshots = zfs._parse_snapshots()
    # comparing .items() because we care about the sorting in the OrderedDict's
    assert snapshots['pool'].items() == expected['pool'].items()
    assert snapshots['pool/fs'].items() == expected['pool/fs'].items()


@pytest.mark.parametrize("fs_name, expected", [
    # (dataset [(snapshot_name, parent_snapshot_name), ...])
    ('pool/fs', [('pool/fs@snap_0', None),
                 ('pool/fs@snap_1_f', 'pool/fs@snap_0'),
                 ('pool/fs@snap_2', 'pool/fs@snap_1_f'),
                 ('pool/fs@snap_3', 'pool/fs@snap_2'),
                 ('pool/fs@snap_8', 'pool/fs@snap_3'),
                 ('pool/fs@snap_9', 'pool/fs@snap_8')]),
    ('pool', [('pool@snap_p1', None),
              ('pool@snap_p2', 'pool@snap_p1')]),
])
def test_zfs_list(fs_name, expected):
    zfs = FakeZFSManager(fs_name=fs_name, snapshot_prefix='snap_')
    actual = [
        (snap.name, snap.parent.name if snap.parent else None)
        for snap in zfs.list()]
    assert actual == expected


class FakeCommandExecutor(CommandExecutor):
    has_pv = False  # disable pv for consistent test output

    def __init__(self, *a, **kwa):
        super(FakeCommandExecutor, self).__init__(*a, **kwa)
        self._called_commands = []
        self._expected = "\nsize 1234"  # currently we only need the output of 'zfs send -nvP' in the tests

    def shell(self, cmd, dry_run=None, capture=None):  # pylint: disable=arguments-differ
        self._called_commands.append(cmd)
        return self._expected


@pytest.fixture
def pair_manager(s3_manager):
    zfs_manager = FakeZFSManager(fs_name='pool/fs', snapshot_prefix='snap_')
    fake_cmd = FakeCommandExecutor()
    return PairManager(s3_manager, zfs_manager, command_executor=fake_cmd)


def test_pair_list(pair_manager):
    pairs = pair_manager.list()
    name = lambda snap: snap.name if snap is not None else None
    names = [
        (name(s3_snap), name(z_snap))
        for (s3_snap, z_snap) in pairs]
    expected = [
        ('pool/fs@snap_0', 'pool/fs@snap_0'),
        ('pool/fs@snap_1_f', 'pool/fs@snap_1_f'),
        ('pool/fs@snap_2', 'pool/fs@snap_2'),
        ('pool/fs@snap_3', 'pool/fs@snap_3'),
        (None, 'pool/fs@snap_8'),  # snap_8 doesn't exist in the s3 fixture
        (None, 'pool/fs@snap_9'),  # snap_9 doesn't exist in the s3 fixture
        # s3-only snapshot pairs are listed last
        ('pool/fs@snap_4_mp', None),
        ('pool/fs@snap_5', None),
        ('pool/fs@snap_6_cycle', None),
        ('pool/fs@snap_7_cycle', None),
    ]
    assert names == expected


def test_backup_latest_full(pair_manager):
    pair_manager.backup_full()
    expected = [
        "zfs send -nvP 'pool/fs@snap_9'",
        ("zfs send 'pool/fs@snap_9' | "
         "pput --estimated 1234 --meta is_full=true {}pool/fs@snap_9")]
    assert pair_manager._cmd._called_commands == [
        e.format(FakeBucket.rand_prefix)
        for e in expected]


def test_backup_full(pair_manager):
    pair_manager.backup_full('pool/fs@snap_3')
    expected = [
        "zfs send -nvP 'pool/fs@snap_3'",
        ("zfs send 'pool/fs@snap_3' | "
         "pput --estimated 1234 --meta is_full=true {}pool/fs@snap_3")]
    assert pair_manager._cmd._called_commands == [
        e.format(FakeBucket.rand_prefix)
        for e in expected]


def test_backup_incremental_latest(pair_manager):
    pair_manager.backup_incremental()
    # snap_8 and snap_9 exist locally but not in s3
    # snap_8 comes after snap_3
    zfs_list = [
        "zfs send -nvP -i 'pool/fs@snap_3' 'pool/fs@snap_8'",
        ("zfs send -i 'pool/fs@snap_3' 'pool/fs@snap_8' | "
         "pput --estimated 1234 --meta parent=pool/fs@snap_3 {}pool/fs@snap_8"),
        "zfs send -nvP -i 'pool/fs@snap_8' 'pool/fs@snap_9'",
        ("zfs send -i 'pool/fs@snap_8' 'pool/fs@snap_9' | "
         "pput --estimated 1234 --meta parent=pool/fs@snap_8 {}pool/fs@snap_9")
    ]
    expected = [e.format(FakeBucket.rand_prefix) for e in zfs_list]
    assert pair_manager._cmd._called_commands == expected


def test_backup_incremental_missing_parent(s3_manager):
    expected = (
        'pool@p1\t0\t19K\t-\t19K\n'
        'pool/fs@snap_1_f\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_4_mp\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_5\t10.0M\t10.0M\t-\t10.0M\n'
    )
    zfs_manager = FakeZFSManager(fs_name='pool/fs', expected=expected, snapshot_prefix='snap_')
    fake_cmd = FakeCommandExecutor()
    pair_manager = PairManager(s3_manager, zfs_manager, command_executor=fake_cmd)
    with pytest.raises(IntegrityError) as excp_info:
        pair_manager.backup_incremental()
    assert excp_info.value.message == \
        "Broken snapshot detected pool/fs@snap_5, reason: 'parent broken'"
    assert fake_cmd._called_commands == []


def test_backup_incremental_cycle(s3_manager):
    zfs_list = (
        'pool@p1\t0\t19K\t-\t19K\n'
        'pool/fs@snap_1_f\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_2\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_3\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_6_cycle\t10.0M\t10.0M\t-\t10.0M\n'  # these have bad metadata on s3
        'pool/fs@snap_7_cycle\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_8\t10.0M\t10.0M\t-\t10.0M\n'
    )
    zfs_manager = FakeZFSManager(fs_name='pool/fs', expected=zfs_list, snapshot_prefix='snap_')
    fake_cmd = FakeCommandExecutor()
    pair_manager = PairManager(s3_manager, zfs_manager, command_executor=fake_cmd)
    with pytest.raises(IntegrityError) as excp_info:
        pair_manager.backup_incremental()
    assert excp_info.value.message == \
        "Broken snapshot detected pool/fs@snap_7_cycle, reason: 'cycle detected'"
    assert fake_cmd._called_commands == []


def test_restore_full(s3_manager):
    """Test full restore on empty zfs dataset"""
    zfs_list = 'pool@p1\t0\t19K\t-\t19K\n'  # we have no pool/fs snapshots locally
    zfs_manager = FakeZFSManager(fs_name='pool/fs', expected=zfs_list, snapshot_prefix='snap_')
    fake_cmd = FakeCommandExecutor()
    pair_manager = PairManager(s3_manager, zfs_manager, command_executor=fake_cmd)
    pair_manager.restore('pool/fs@snap_1_f')
    expected = "z3_get {}pool/fs@snap_1_f | zfs recv pool/fs@snap_1_f".format(
        FakeBucket.rand_prefix)
    assert fake_cmd._called_commands == [expected]


def test_restore_incremental_empty_dataset(s3_manager):
    """Tests incremental restore on a zfs dataset with no snapshots"""
    zfs_list = 'pool@p1\t0\t19K\t-\t19K\n'  # we have no pool/fs snapshots locally
    zfs_manager = FakeZFSManager(fs_name='pool/fs', expected=zfs_list, snapshot_prefix='snap_')
    fake_cmd = FakeCommandExecutor()
    pair_manager = PairManager(s3_manager, zfs_manager, command_executor=fake_cmd)
    pair_manager.restore('pool/fs@snap_3')  # ask for an incremental snapshot
    # all incremental snapshots until we hit a full snapshot are expected
    expected = [
        "z3_get {}pool/fs@snap_1_f | zfs recv pool/fs@snap_1_f",
        "z3_get {}pool/fs@snap_2 | zfs recv pool/fs@snap_2",
        "z3_get {}pool/fs@snap_3 | zfs recv pool/fs@snap_3",
    ]
    expected = [e.format(FakeBucket.rand_prefix) for e in expected]
    assert fake_cmd._called_commands == expected


def test_restore_incremental(s3_manager):
    """Tests incremental restore on a zfs dataset with existing snapshots"""
    zfs_list = (
        'pool@p1\t0\t19K\t-\t19K\n'  # we have no pool/fs snapshots locally
        'pool/fs@snap_1_f\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_2\t10.0M\t10.0M\t-\t10.0M\n'
    )
    zfs_manager = FakeZFSManager(fs_name='pool/fs', expected=zfs_list, snapshot_prefix='snap_')
    fake_cmd = FakeCommandExecutor()
    pair_manager = PairManager(s3_manager, zfs_manager, command_executor=fake_cmd)
    pair_manager.restore('pool/fs@snap_3')  # ask for an incremental snapshot
    # all incremental snapshots until we hit a full snapshot are expected
    expected = [
        "z3_get {}pool/fs@snap_3 | zfs recv pool/fs@snap_3",
    ]
    expected = [e.format(FakeBucket.rand_prefix) for e in expected]
    assert fake_cmd._called_commands == expected


def test_restore_broken(s3_manager):
    """Tests restoring a broken snapshot raises integrity error"""
    zfs_list = (
        'pool@p1\t0\t19K\t-\t19K\n'  # we have no pool/fs snapshots locally
        'pool/fs@snap_1_f\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_2\t10.0M\t10.0M\t-\t10.0M\n'
    )
    zfs_manager = FakeZFSManager(fs_name='pool/fs', expected=zfs_list, snapshot_prefix='snap_')
    fake_cmd = FakeCommandExecutor()
    pair_manager = PairManager(s3_manager, zfs_manager, command_executor=fake_cmd)
    with pytest.raises(IntegrityError) as excp_info:
        pair_manager.restore('pool/fs@snap_4_mp')
    assert excp_info.value.message == \
        "Broken snapshot detected pool/fs@snap_4_mp, reason: 'missing parent'"


def test_restore_noop(s3_manager):
    """Test restore does nothing when snapshot already exist locally"""
    expected = (
        'pool@p1\t0\t19K\t-\t19K\n'  # we have no pool/fs snapshots locally
        'pool/fs@snap_1_f\t10.0M\t10.0M\t-\t10.0M\n'
        'pool/fs@snap_2\t10.0M\t10.0M\t-\t10.0M\n'
    )
    zfs_manager = FakeZFSManager(fs_name='pool/fs', expected=expected, snapshot_prefix='snap_')
    fake_cmd = FakeCommandExecutor()
    pair_manager = PairManager(s3_manager, zfs_manager, command_executor=fake_cmd)
    pair_manager.restore('pool/fs@snap_2')
    assert fake_cmd._called_commands == []



import argparse
from datetime import datetime

import boto

from z3.config import get_config


def cleanup_multipart(bucket, max_days=1, dry_run=False):
    max_age_seconds = max_days * 24 * 3600
    now = datetime.utcnow()
    fmt = "{} | {:30} | {:20}"
    print fmt.format("A", "key", "initiated")
    for multi in bucket.list_multipart_uploads():
        delta = now-boto.utils.parse_ts(multi.initiated)
        if delta.total_seconds() >= max_age_seconds:
            print fmt.format("X", multi.key_name, multi.initiated)
            if not dry_run:
                multi.cancel_upload()
        else:
            print fmt.format(" ", multi.key_name, multi.initiated)


def main():
    cfg = get_config()
    parser = argparse.ArgumentParser(
        description='Cleanup hanging multipart s3 uploads',
    )
    parser.add_argument('--max-age',
                        dest='max_days',
                        default=1,
                        type=int,
                        help='maximum age in days')
    parser.add_argument('--dry',
                        dest='dry_run',
                        action='store_true',
                        help='Don\'t cancel any upload')
    args = parser.parse_args()
    bucket = boto.connect_s3(
        cfg['S3_KEY_ID'], cfg['S3_SECRET']).get_bucket(cfg['BUCKET'])
    cleanup_multipart(
        bucket,
        max_days=args.max_days,
        dry_run=args.dry_run,
    )


if __name__ == '__main__':
    main()



"""Multipart parallel s3 upload.

usage
pput bucket_name/filename
"""

from Queue import Queue
from cStringIO import StringIO
from collections import namedtuple
from threading import Thread
import argparse
import binascii
import functools
import hashlib
import logging
import json
import os
import sys

import boto.s3.multipart

from z3.config import get_config


Result = namedtuple('Result', ['success', 'traceback', 'index', 'md5'])
CFG = get_config()
VERB_QUIET = 0
VERB_NORMAL = 1
VERB_PROGRESS = 2


def multipart_etag(digests):
    """
    Computes etag for multipart uploads
    :type digests: list of hex-encoded md5 sums (string)
    :param digests: The list of digests for each individual chunk.

    :rtype: string
    :returns: The etag computed from the individual chunks.
    """
    etag = hashlib.md5()
    count = 0
    for dig in digests:
        count += 1
        etag.update(binascii.a2b_hex(dig))
    return '"{}-{}"'.format(etag.hexdigest(), count)


def parse_size(size):
    if isinstance(size, (int, long)):
        return size
    size = size.strip().upper()
    last = size[-1]
    if last == 'T':
        return int(size[:-1]) * 1024 * 1024 * 1024 * 1024
    if last == 'G':
        return int(size[:-1]) * 1024 * 1024 * 1024
    if last == 'M':
        return int(size[:-1]) * 1024 * 1024
    if last == 'K':
        return int(size[:-1]) * 1024
    return int(size)


class StreamHandler(object):
    def __init__(self, input_stream, chunk_size=5*1024*1024):
        self.input_stream = input_stream
        self.chunk_size = chunk_size
        self._partial_chunk = ""
        self._eof_reached = False

    @property
    def finished(self):
        return self._eof_reached and len(self._partial_chunk) == 0

    def get_chunk(self):
        """Return complete chunks or None if EOF reached"""
        while not self._eof_reached:
            read = self.input_stream.read(self.chunk_size - len(self._partial_chunk))
            if len(read) == 0:
                self._eof_reached = True
            self._partial_chunk += read
            if len(self._partial_chunk) == self.chunk_size or self._eof_reached:
                chunk = self._partial_chunk
                self._partial_chunk = ""
                return chunk
            # else:
            #     print "partial", len(self._partial_chunk)


def retry(times=int(CFG['MAX_RETRIES'])):
    def decorator(func):
        @functools.wraps(func)
        def wrapped(*a, **kwa):
            for attempt in xrange(1, times+1):
                try:
                    return func(*a, **kwa)
                except: # pylint: disable=bare-except
                    if attempt >= times:
                        raise
                    logging.exception('Failed to upload part attempt {} of {}'.format(
                        attempt, times))
        return wrapped
    return decorator


class WorkerCrashed(Exception):
    pass


class UploadWorker(object):
    def __init__(self, bucket, multipart, inbox, outbox):
        self.bucket = bucket
        self.inbox = inbox
        self.outbox = outbox
        self.multipart = multipart
        self._thread = None
        self.log = logging.getLogger('UploadWorker')

    @retry()
    def upload_part(self, index, chunk):
        part = boto.s3.multipart.MultiPartUpload(self.bucket)
        part.id = self.multipart.id
        part.key_name = self.multipart.key_name
        return part.upload_part_from_file(
            StringIO(chunk), index, replace=True).md5

    def start(self):
        self._thread = Thread(target=self.main_loop)
        self._thread.daemon = True
        self._thread.start()
        return self

    def is_alive(self):
        return self._thread.is_alive()

    def main_loop(self):
        while True:
            index, chunk = self.inbox.get()
            md5 = self.upload_part(index, chunk)
            # print "worker loop i:{} md5:{}".format(index, md5)
            self.outbox.put(Result(
                success=True,
                md5=md5,
                traceback=None,
                index=index,
            ))


class UploadException(Exception):
    pass


class UploadSupervisor(object):
    '''Reads chunks and dispatches them to UploadWorkers'''

    def __init__(self, stream_handler, name, bucket, headers=None, verbosity=1):
        self.stream_handler = stream_handler
        self.name = name
        self.bucket = bucket
        self.inbox = None
        self.outbox = None
        self.multipart = None
        self.results = []  # beware s3 multipart indexes are 1 based
        self._pending_chunks = 0
        self._verbosity = verbosity
        self._workers = None
        self._headers = headers

    def _start_workers(self, concurrency, worker_class):
        work_queue = Queue(maxsize=concurrency)
        result_queue = Queue()
        self.outbox = work_queue
        self.inbox = result_queue
        workers = [
            worker_class(
                bucket=self.bucket,
                multipart=self.multipart,
                inbox=work_queue,
                outbox=result_queue,
            ).start()
            for _ in xrange(concurrency)]
        return workers

    def _begin_upload(self):
        if self.multipart is not None:
            raise AssertionError("multipart upload already started")
        headers = {
            "x-amz-acl": "bucket-owner-full-control",
            "x-amz-storage-class": 'STANDARD_IA',
        }
        if self._headers:
            headers.update(self._headers)
        self.multipart = self.bucket.initiate_multipart_upload(self.name, headers=headers)

    def _finish_upload(self):
        if len(self.results) == 0:
            self.multipart.cancel_upload()
            raise UploadException("Error: Can't upload zero bytes!")
        return self.multipart.complete_upload()

    def _handle_result(self):
        """Process one result. Block untill one is available
        """
        result = self.inbox.get()
        if result.success:
            if self._verbosity >= VERB_PROGRESS:
                sys.stderr.write("\nuploaded chunk {} \n".format(result.index))
            self.results.append((result.index, result.md5))
            self._pending_chunks -= 1
        else:
            raise result.traceback

    def _handle_results(self):
        """Process any available result
        Doesn't block.
        """
        while not self.inbox.empty():
            self._handle_result()

    def _send_chunk(self, index, chunk):
        """Send the current chunk to the workers for processing.
        Called when the _partial_chunk is complete.

        Blocks when the outbox is full.
        """
        self._pending_chunks += 1
        self.outbox.put((index, chunk))

    def _check_workers(self):
        """Check workers are alive, raise exception if any is dead."""
        for worker in self._workers:
            if not worker.is_alive():
                raise WorkerCrashed()

    def main_loop(self, concurrency=4, worker_class=UploadWorker):
        chunk_index = 0
        self._begin_upload()
        self._workers = self._start_workers(concurrency, worker_class=worker_class)
        while self._pending_chunks or not self.stream_handler.finished:
            self._check_workers()  # raise exception and stop everything if any worker has crashed
            # print "main_loop p:{} o:{} i:{}".format(
            #     self._pending_chunks, self.outbox.qsize(), self.inbox.qsize())
            # consume results first as this is a quick operation
            self._handle_results()
            chunk = self.stream_handler.get_chunk()
            if chunk:
                # s3 multipart index is 1 based, increment before sending
                chunk_index += 1
                self._send_chunk(chunk_index, chunk)
        self._finish_upload()
        self.results.sort()
        return multipart_etag(r[1] for r in self.results)


def parse_metadata(metadata):
    headers = {}
    for meta in metadata:
        try:
            key, val = meta.split('=', 1)
        except ValueError:
            sys.stderr.write(
                "malformed metadata '{}'; should be key=value\n".format(meta))
            sys.exit(1)
        headers['x-amz-meta-' + key] = val
    return headers


def optimize_chunksize(estimated):
    max_parts = 9999  # S3 requires part indexes to be between 1 and 10000
    # part size has to be at least 5MB
    estimated = estimated * 1.05  # just to be on the safe side overesimate the total size to upload
    min_part_size = max(estimated / max_parts, 5*1024*1024)
    return int(min_part_size)


def main():
    parser = argparse.ArgumentParser(
        description='Read data from stdin and upload it to s3',
        epilog=('All optional args have a configurable default. '
                'Order of precedence is command line args then '
                'environment variables then user config ~/.z3.cfg'
                ' then default config.'),
    )
    parser.add_argument('name', help='name of S3 key')
    chunk_group = parser.add_mutually_exclusive_group()
    chunk_group.add_argument('-s', '--chunk-size',
                             dest='chunk_size',
                             default=CFG['CHUNK_SIZE'],
                             help='multipart chunk size, eg: 10M, 1G')
    chunk_group.add_argument('--estimated',
                             help='Estimated upload size')
    parser.add_argument('--file-descriptor',
                        dest='file_descriptor',
                        type=int,
                        help=('read data from this fd instead of stdin; '
                              'useful if you want an [i]pdb session to use stdin\n'
                              '`pput --file-descriptor 3 3<./file`'))
    parser.add_argument('--concurrency',
                        dest='concurrency',
                        type=int,
                        default=int(CFG['CONCURRENCY']),
                        help='number of worker threads to use')
    parser.add_argument('--metadata',
                        action='append',
                        dest='metadata',
                        default=list(),
                        help='Metatada in key=value format')
    quiet_group = parser.add_mutually_exclusive_group()
    quiet_group.add_argument('--progress',
                             dest='progress',
                             action='store_true',
                             help=('show progress report'))
    quiet_group.add_argument('--quiet',
                             dest='quiet',
                             action='store_true',
                             help=('don\'t emit any output at all'))
    args = parser.parse_args()

    input_fd = os.fdopen(args.file_descriptor, 'r') if args.file_descriptor else sys.stdin
    if args.estimated is not None:
        chunk_size = optimize_chunksize(parse_size(args.estimated))
    else:
        chunk_size = parse_size(args.chunk_size)
    stream_handler = StreamHandler(input_fd, chunk_size=chunk_size)
    bucket = boto.connect_s3(
        CFG['S3_KEY_ID'], CFG['S3_SECRET']).get_bucket(CFG['BUCKET'])
    verbosity = 0 if args.quiet else 1 + int(args.progress)  # 0 totally silent, 1 default, 2 show progress
    sup = UploadSupervisor(
        stream_handler,
        args.name,
        bucket=bucket,
        verbosity=verbosity,
        headers=parse_metadata(args.metadata),
    )
    if verbosity >= VERB_NORMAL:
        sys.stderr.write("starting upload to {}/{} with chunksize {}M using {} workers\n".format(
            CFG['BUCKET'], args.name, (chunk_size/(1024*1024.0)), args.concurrency))
    try:
        etag = sup.main_loop(concurrency=args.concurrency)
    except UploadException as excp:
        sys.stderr.write("{}\n".format(excp))
        return 1
    if verbosity >= VERB_NORMAL:
        print json.dumps({'status': 'success', 'etag': etag})


if __name__ == '__main__':
    main()



# pylint: disable=redefined-outer-name,protected-access
import pytest

from z3.ssh_sync import snapshots_to_send, sync_snapshots


ALL = ['S_0', 'S_1', 'S_2', 'S_3', 'S_4']


HAPPY_PATH = dict(
    # empty remote, send latest; no base snapshot
    empty_remote=((ALL, []), (None, 'S_4')),

    # first snapshot common to both, send latest, based on S_0
    first_sent=((ALL, ['S_0']), ('S_0', 'S_4')),

    # send latest, based on S_3
    latest=((ALL, ['S_1', 'S_3']), ('S_3', 'S_4')),

    # nothing to do
    noop=((ALL[1:], ALL), ('S_4', 'S_4')),
)


@pytest.mark.parametrize("pair, expected", HAPPY_PATH.values(), ids=HAPPY_PATH.keys())
def test_snapshots_to_send(pair, expected):
    local, remote = pair
    assert snapshots_to_send(local, remote) == expected


ERRORS = dict(
    empty_local=(([], []),
                 "No snapshots exist locally!"),
    no_common_snapshots=((["S_0"], ["S_10"]),
                         "Latest snapshot on destination doesn't exist on source!"),
)


@pytest.mark.parametrize('pair, err_msg', ERRORS.values(), ids=ERRORS.keys())
def test_snapshots_to_send_error(pair, err_msg):
    local, remote = pair
    with pytest.raises(AssertionError) as err:
        snapshots_to_send(local, remote)
    assert err_msg == err.value.message


PULL_HAPPY_PATH = dict(
    incremental=(
        ('S_0', 'S_4'),
        ("ssh example.com -C 'sudo zfs send -I remote/fs@S_0 remote/fs@S_4'",
         'mbuffer -s 128k -m 200m -q | zfs recv -d local/fs')),
    empty_target=(
        (None, 'S_4'),
        ("ssh example.com -C 'sudo zfs send remote/fs@S_4'",
         'mbuffer -s 128k -m 200m -q | zfs recv -d local/fs')),
    noop=(
        ('S_4', 'S_4'),
        None),
)


@pytest.mark.parametrize('pair, expected', PULL_HAPPY_PATH.values(), ids=PULL_HAPPY_PATH.keys())
def test_pull_command(pair, expected):
    commands = sync_snapshots(
        pair,
        local_fs='local/fs',
        remote_fs='remote/fs',
        remote_addr='example.com',
        pull=True,
        dry_run=False,
    )
    assert commands == expected


PUSH_HAPPY_PATH = dict(
    incremental=(
        ('S_0', 'S_4'),
        ("zfs send -I local/fs@S_0 local/fs@S_4",
         "ssh example.com -C 'mbuffer -s 128k -m 200m -q | sudo zfs recv -d remote/fs'")),
    empty_target=(
        (None, 'S_4'),
        ("zfs send local/fs@S_4",
         "ssh example.com -C 'mbuffer -s 128k -m 200m -q | sudo zfs recv -d remote/fs'")),
    noop=(
        ('S_4', 'S_4'),
        None),
)


@pytest.mark.parametrize('pair, expected', PUSH_HAPPY_PATH.values(), ids=PUSH_HAPPY_PATH.keys())
def test_push_command(pair, expected):
    commands = sync_snapshots(
        pair,
        local_fs='local/fs',
        remote_fs='remote/fs',
        remote_addr='example.com',
        pull=False,
        dry_run=False,
    )
    assert commands == expected



from __future__ import print_function

import argparse
import subprocess
import sys

from z3.config import get_config
from z3.snap import ZFSSnapshotManager, CommandExecutor


quiet = False


class RemoteZFSSnapshotManager(ZFSSnapshotManager):
    def __init__(self, remote_addr, *a, **kwa):
        super(RemoteZFSSnapshotManager, self).__init__(*a, **kwa)
        self.remote_addr = remote_addr

    def _list_snapshots(self):
        return subprocess.check_output(
            ['ssh', self.remote_addr, '-C',
             'sudo zfs list -Ht snap -o name,used,refer,mountpoint,written'])


def snapshots_to_send(source_snaps, dest_snaps):
    """return pair of snapshots"""
    if len(source_snaps) == 0:
        raise AssertionError("No snapshots exist locally!")
    if len(dest_snaps) == 0:
        # nothing on the remote side, send everything
        return None, source_snaps[-1]
    last_remote = dest_snaps[-1]
    for snap in reversed(source_snaps):
        if snap == last_remote:
            # found a common snapshot
            return last_remote, source_snaps[-1]
    # sys.stderr.write("source:'{}', dest:'{}'".format(source_snaps, dest_snaps))
    raise AssertionError("Latest snapshot on destination doesn't exist on source!")


def prepare_commands(from_snap, to_snap, filesystem, dry_run=False):
    if from_snap == to_snap:
        if not quiet:
            print("Nothing to do here.")
        return
    if from_snap is None:
        send_cmd = "zfs send {}".format(to_snap)
    else:
        send_cmd = "zfs send -I {} {}".format(from_snap, to_snap)
    dry = 'nv' if dry_run else ''
    recv_cmd = "zfs recv -d{dry} {}".format(
        filesystem, dry=dry)
    return send_cmd, recv_cmd


def send_snapshots(send_cmd, recv_cmd, remote_addr):
    recv_cmd = "ssh {} -C 'mbuffer -s 128k -m 200m -q | sudo {}'".format(
        remote_addr, recv_cmd)
    return send_cmd, recv_cmd


def pull_snapshots(send_cmd, recv_cmd, remote_addr):
    send_cmd = "ssh {} -C 'sudo {}'".format(
        remote_addr, send_cmd)
    recv_cmd = "mbuffer -s 128k -m 200m -q | {}".format(recv_cmd)
    return send_cmd, recv_cmd


def sync_snapshots(pair, local_fs, remote_fs, remote_addr, pull, dry_run):
    from_snap, to_snap = pair
    target_fs = local_fs if pull else remote_fs
    source_fs = remote_fs if pull else local_fs
    from_snap = "{}@{}".format(source_fs, from_snap) if from_snap is not None else None
    to_snap = "{}@{}".format(source_fs, to_snap)
    cmd_pair = prepare_commands(
        from_snap,
        to_snap,
        filesystem=target_fs,
        dry_run=dry_run,
    )
    if cmd_pair is None:
        return
    send_cmd, recv_cmd = cmd_pair
    if pull:
        return pull_snapshots(send_cmd, recv_cmd, remote_addr)
    else:
        return send_snapshots(send_cmd, recv_cmd, remote_addr)


def main():
    global quiet
    cfg = get_config()
    parser = argparse.ArgumentParser(
        description='send z3 snapshots over ssh',
    )
    parser.add_argument('--filesystem', '--dataset',
                        dest='filesystem',
                        default=cfg.get('FILESYSTEM'),
                        help='the zfs dataset/filesystem to operate on')
    parser.add_argument('--remote-filesystem', '--remote-dataset',
                        dest='remote_filesystem',
                        default=None,
                        help='the target zfs dataset/filesystem to send snapshots to')
    parser.add_argument('--snapshot-prefix',
                        dest='snapshot_prefix',
                        default=cfg.get('SNAPSHOT_PREFIX', 'zfs-auto-snap:daily'),
                        help='only operate on snapshots that start with this prefix')
    parser.add_argument('--pull',
                        dest='pull',
                        default=False,
                        action='store_true',
                        help='pull snapshots from remote')
    parser.add_argument('--quiet',
                        dest='quiet',
                        default=False,
                        action='store_true',
                        help='suppress output')
    parser.add_argument('--dry-run',
                        dest='dry_run',
                        default=False,
                        action='store_true',
                        help='call zfs recv with -nv flags to test if snapshot can be sent')
    parser.add_argument('remote', help='hostname/address of remote server')
    args = parser.parse_args()
    quiet = args.quiet
    local_mgr = ZFSSnapshotManager(args.filesystem, args.snapshot_prefix)
    remote_fs = args.remote_filesystem or args.filesystem
    remote_mgr = RemoteZFSSnapshotManager(args.remote, remote_fs, args.snapshot_prefix)
    local_snaps = [s.name[len(args.filesystem)+1:]  # strip fs name
                   for s in local_mgr.list()]
    remote_snaps = [s.name[len(remote_fs)+1:]  # strip fs name
                    for s in remote_mgr.list()]
    if args.pull:
        pair = snapshots_to_send(source_snaps=remote_snaps, dest_snaps=local_snaps)
    else:
        pair = snapshots_to_send(source_snaps=local_snaps, dest_snaps=remote_snaps)
    cmd_pair = sync_snapshots(
        pair, args.filesystem, remote_fs, args.remote, args.pull, dry_run=args.dry_run)
    if cmd_pair is None:
        return
    send_cmd, recv_cmd = cmd_pair
    executor = CommandExecutor()
    executor.pipe(send_cmd, recv_cmd, quiet=quiet)



if __name__ == '__main__':
    main()



import argparse
import sys

import boto.s3

from z3.config import get_config


def download(bucket, name):
    key = bucket.get_key(name)
    key.get_contents_to_file(sys.stdout)


def main():
    cfg = get_config()
    parser = argparse.ArgumentParser(
        description='Read a key from s3 and write the content to stdout',
    )
    parser.add_argument('name', help='name of S3 key')
    args = parser.parse_args()
    bucket = boto.connect_s3(
        cfg['S3_KEY_ID'], cfg['S3_SECRET']).get_bucket(cfg['BUCKET'])
    download(bucket, args.name)

if __name__ == '__main__':
    main()



import ConfigParser
import os
import os.path

import z3


_settings = None


class OnionDict(object):
    """Wrapps multiple dictionaries. Tries to read data from each dict
    in turn.
    Used to implement a fallback mechanism.
    """
    def __init__(self, *args):
        self.__dictionaries = args

    def __getitem__(self, key):
        for d in self.__dictionaries:
            if key in d:
                return d[key]
        raise KeyError(key)

    def __contains__(self, key):
        for d in self.__dictionaries:
            if key in d:
                return True
        return False

    def get(self, key, default=None):
        if key in self:
            return self[key]
        else:
            return default


def get_config():
    global _settings
    if _settings is None:
        _config = ConfigParser.ConfigParser()
        default = os.path.join(z3.__path__[0], "z3.conf")
        _config.read(default)
        _config.read("/etc/z3_backup/z3.conf")
        _settings = OnionDict(
            os.environ,  # env variables take precedence
            dict((k.upper(), v) for k, v in _config.items("main"))
        )
    return _settings



from __future__ import print_function

import argparse
import functools
import logging
import operator
import os
import subprocess
from collections import OrderedDict

import boto

from z3.config import get_config


quiet = False


def cached(func):
    @functools.wraps(func)
    def wrapper(self, *a, **kwa):
        # ARGUMENTS AREN'T PART OF CACHE KEY
        # that's safe in this case, but might cause surprising bugs if you reuse it as is
        cache_key = func.__name__ + '_cached_value'
        if not hasattr(self, cache_key):
            val = func(self, *a, **kwa)
            setattr(self, cache_key, val)
        return getattr(self, cache_key)
    return wrapper


class IntegrityError(Exception):
    pass


class S3Snapshot(object):
    CYCLE = 'cycle detected'
    MISSING_PARENT = 'missing parent'
    PARENT_BROKEN = 'parent broken'

    def __init__(self, name, metadata, manager):
        self.name = name
        self._metadata = metadata
        self._mgr = manager
        self._reason_broken = None

    def __repr__(self):
        if self.is_full:
            return "<Snapshot {} [full]>".format(self.name)
        else:
            return "<Snapshot {} [{}]>".format(self.name, self.parent_name)

    @property
    def is_full(self):
        return self._metadata.get('is_full') == 'true'

    @property
    def parent(self):
        parent_name = self._metadata.get('parent')
        return self._mgr.get(parent_name)

    @property
    def parent_name(self):
        return self._metadata.get("parent")

    @cached
    def _is_healthy(self, visited=frozenset()):
        if self.is_full:
            return True
        if self in visited:
            self._reason_broken = self.CYCLE
            return False  # we ended up with a cycle, abort
        if self.parent is None:
            self._reason_broken = self.MISSING_PARENT
            return False  # missing parent
        if not self.parent._is_healthy(visited.union([self])):
            if self.parent.reason_broken == self.CYCLE:
                self._reason_broken = self.CYCLE
            else:
                self._reason_broken = self.PARENT_BROKEN
            return False
        return True

    @property
    def is_healthy(self):
        return self._is_healthy()

    @property
    def reason_broken(self):
        if self.is_healthy:
            return
        return self._reason_broken


class S3SnapshotManager(object):
    def __init__(self, bucket, s3_prefix, snapshot_prefix):
        self.bucket = bucket
        self.s3_prefix = s3_prefix.rstrip('/') + '/'  # make sure we always have a trailing /
        self.snapshot_prefix = snapshot_prefix

    @property
    @cached
    def _snapshots(self):
        prefix = os.path.join(self.s3_prefix, self.snapshot_prefix)
        snapshots = {}
        strip_chars = len(self.s3_prefix)
        for key in self.bucket.list(prefix):
            # print key.key, prefix
            key = self.bucket.get_key(key.key)
            name = key.key[strip_chars:]
            snapshots[name] = S3Snapshot(name, metadata=key.metadata, manager=self)
        # from pprint import pprint
        # pprint(snapshots)
        return snapshots

    def list(self):
        return sorted(self._snapshots.values(), key=operator.attrgetter('name'))

    def get(self, name):
        return self._snapshots.get(name)


class ZFSSnapshot(object):
    def __init__(self, name, metadata, parent=None, manager=None):
        self.name = name
        self.parent = parent

    def __repr__(self):
        return "<Snapshot {} [{}]>".format(self.name, self.parent.name if self.parent else '')


class ZFSSnapshotManager(object):
    def __init__(self, fs_name, snapshot_prefix):
        self._fs_name = fs_name
        self._snapshot_prefix = snapshot_prefix
        self._sorted = None

    def _list_snapshots(self):
        # This is overridden in tests
        # see FakeZFSManager
        return subprocess.check_output(
            ['zfs', 'list', '-Ht', 'snap', '-o',
             'name,used,refer,mountpoint,written'])

    def _parse_snapshots(self):
        """Returns all snapshots grouped by filesystem, a dict of OrderedDict's
        The order of snapshots matters when determining parents for incremental send,
        so it's preserved.
        Data is indexed by filesystem then for each filesystem we have an OrderedDict
        of snapshots.
        """
        try:
            snap = self._list_snapshots()
        except OSError as err:
            logging.error("unable to list local snapshots!")
            return {}
        vols = {}
        for line in snap.splitlines():
            name, used, refer, mountpoint, written = line.split('\t')
            vol_name, snap_name = name.split('@', 1)
            snapshots = vols.setdefault(vol_name, OrderedDict())
            snapshots[snap_name] = {
                'name': name,
                'used': used,
                'refer': refer,
                'mountpoint': mountpoint,
                'written': written,
            }
        return vols

    def _build_snapshots(self, fs_name):
        snapshots = OrderedDict()
        # for fs_name, fs_snaps in self._parse_snapshots().iteritems():
        fs_snaps = self._parse_snapshots().get(fs_name, {})
        parent = None
        for snap_name, data in fs_snaps.iteritems():
            if not snap_name.startswith(self._snapshot_prefix):
                continue
            full_name = '{}@{}'.format(fs_name, snap_name)
            zfs_snap = ZFSSnapshot(
                full_name,
                metadata=data,
                parent=parent,
                manager=self,
            )
            snapshots[full_name] = zfs_snap
            parent = zfs_snap
        return snapshots

    @property
    @cached
    def _snapshots(self):
        return self._build_snapshots(self._fs_name)

    def list(self):
        return self._snapshots.values()

    def get_latest(self):
        return self._snapshots.values()[-1]

    def get(self, name):
        return self._snapshots.get(name)


class CommandExecutor(object):
    @staticmethod
    def shell(cmd, dry_run=False, capture=False):
        if dry_run:
            print(cmd)
        else:
            if capture:
                return subprocess.check_output(
                    cmd, shell=True, stderr=subprocess.STDOUT)
            else:
                return subprocess.check_call(
                    cmd, shell=True)

    @property
    @cached
    def has_pv(self):
        return subprocess.call(
            ['which', 'pv'],
            stderr=subprocess.STDOUT, stdout=subprocess.PIPE) == 0

    def pipe(self, cmd1, cmd2, quiet=False, **kwa):
        """Executes commands"""
        if self.has_pv and not quiet:
            return self.shell("{} | pv | {}".format(cmd1, cmd2), **kwa)
        else:
            return self.shell("{} | {}".format(cmd1, cmd2), **kwa)


class PairManager(object):
    def __init__(self, s3_manager, zfs_manager, command_executor=None):
        self.s3_manager = s3_manager
        self.zfs_manager = zfs_manager
        self._cmd = command_executor or CommandExecutor()

    def list(self):
        pairs = []
        seen = set([])
        for z_snap in self.zfs_manager.list():
            seen.add(z_snap.name)
            pairs.append(
                (self.s3_manager.get(z_snap.name), z_snap))
        for s3_snap in self.s3_manager.list():
            if s3_snap.name not in seen:
                pairs.append((s3_snap, None))
        return pairs

    def _snapshot_to_backup(self, snap_name):
        if snap_name is None:
            z_snap = self.zfs_manager.get_latest()
        else:
            z_snap = self.zfs_manager.get(snap_name)
            if z_snap is None:
                raise Exception('Failed to get the snapshot {}'.format(snap_name))
        return z_snap

    @staticmethod
    def _parse_estimated_size(output):
        try:
            size_line = [line for line in output.splitlines() if len(line)][-1]
            _, size = size_line.split()
            return int(size)
        except:
            logging.error("failed to parse output '%s'", output)
            raise

    def backup_full(self, snap_name=None, dry_run=False):
        """Do a full backup of a snapshot. By default latest local snapshot"""
        z_snap = self._snapshot_to_backup(snap_name)
        estimated_size = self._parse_estimated_size(
            self._cmd.shell(
                "zfs send -nvP '{}'".format(z_snap.name),
                capture=True))
        self._cmd.pipe(
            "zfs send '{}'".format(z_snap.name),
            "pput --estimated {} --meta is_full=true {}{}".format(
                estimated_size, self.s3_manager.s3_prefix, z_snap.name),
            dry_run=dry_run,
        )

    def backup_incremental(self, snap_name=None, dry_run=False):
        """Uploads named snapshot or latest, along with any other snapshots
        required for an incremental backup.
        """
        z_snap = self._snapshot_to_backup(snap_name)
        to_upload = []
        current = z_snap
        while True:
            s3_snap = self.s3_manager.get(current.name)
            if s3_snap is not None:
                if not s3_snap.is_healthy:
                    # abort everything if we run in to unhealthy snapshots
                    raise IntegrityError(
                        "Broken snapshot detected {}, reason: '{}'".format(
                            s3_snap.name, s3_snap.reason_broken
                        ))
                break
            to_upload.append(current)
            if current.parent is None:
                break
            current = current.parent
        for z_snap in reversed(to_upload):
            estimated_size = self._parse_estimated_size(
                self._cmd.shell(
                    "zfs send -nvP -i '{}' '{}'".format(
                        z_snap.parent.name, z_snap.name),
                    capture=True))
            self._cmd.pipe(
                "zfs send -i '{}' '{}'".format(
                    z_snap.parent.name, z_snap.name),
                "pput --estimated {} --meta parent={} {}{}".format(
                    estimated_size, z_snap.parent.name, self.s3_manager.s3_prefix, z_snap.name),
                dry_run=dry_run,
            )

    def restore(self, snap_name, dry_run=False):
        current_snap = self.s3_manager.get(snap_name)
        if current_snap is None:
            raise Exception('no such snapshot "{}"'.format(snap_name))
        to_restore = []
        while True:
            z_snap = self.zfs_manager.get(current_snap.name)
            if z_snap is not None:
                break
            if not current_snap.is_healthy:
                raise IntegrityError(
                    "Broken snapshot detected {}, reason: '{}'".format(
                        current_snap.name, current_snap.reason_broken
                    ))
            to_restore.append(current_snap)
            if current_snap.is_full:
                break
            else:
                current_snap = current_snap.parent
        for s3_snap in reversed(to_restore):
            self._cmd.pipe(
                "z3_get {}".format(
                    os.path.join(self.s3_manager.s3_prefix, s3_snap.name)),
                "zfs recv {}".format(s3_snap.name),
                dry_run=dry_run,
            )


def _get_widths(widths, line):
    for index, value in enumerate(line):
        widths[index] = max(widths[index], len("{}".format(value)))
    return widths


def _prepare_line(s3_snap, z_snap):
    if s3_snap is None:
        snap_type = 'missing'
        health = '-'
        name = z_snap.name.split('@', 1)[1]
        parent_name = '-'
        local_state = 'ok'
    else:
        snap_type = 'full' if s3_snap.is_full else 'incremental'
        health = s3_snap.reason_broken or 'ok'
        parent_name = '' if s3_snap.is_full else s3_snap.parent_name
        name = s3_snap.name
        local_state = 'ok' if z_snap is not None else 'missing'
    return (name, parent_name, snap_type, health, local_state)


def list_snapshots(bucket, s3_prefix, filesystem, snapshot_prefix):
    print("backup status for {}@{}* on {}/{}".format(
        filesystem, snapshot_prefix, bucket.name, s3_prefix))
    prefix = "{}@{}".format(filesystem, snapshot_prefix)
    pair_manager = PairManager(
        S3SnapshotManager(bucket, s3_prefix=s3_prefix, snapshot_prefix=prefix),
        ZFSSnapshotManager(fs_name=filesystem, snapshot_prefix=snapshot_prefix))
    header = ("NAME", "PARENT", "TYPE", "HEALTH", "LOCAL STATE")
    widths = [len(col) for col in header]
    listing = []
    for s3_snap, z_snap in pair_manager.list():
        line = _prepare_line(s3_snap, z_snap)
        listing.append(line)
        widths = _get_widths(widths, line)
    fmt = " | ".join("{{:{w}}}".format(w=w) for w in widths)
    print(fmt.format(*header))
    for line in listing:
        print(fmt.format(*line))


def do_backup(bucket, s3_prefix, filesystem, snapshot_prefix, full, snapshot, dry):
    prefix = "{}@{}".format(filesystem, snapshot_prefix)
    s3_mgr = S3SnapshotManager(bucket, s3_prefix=s3_prefix, snapshot_prefix=prefix)
    zfs_mgr = ZFSSnapshotManager(fs_name=filesystem, snapshot_prefix=snapshot_prefix)
    pair_manager = PairManager(s3_mgr, zfs_mgr)
    snap_name = "{}@{}".format(filesystem, snapshot) if snapshot else None
    if full is True:
        pair_manager.backup_full(snap_name=snap_name, dry_run=dry)
    else:
        pair_manager.backup_incremental(snap_name=snap_name, dry_run=dry)


def restore(bucket, s3_prefix, filesystem, snapshot_prefix, snapshot, dry):
    prefix = "{}@{}".format(filesystem, snapshot_prefix)
    s3_mgr = S3SnapshotManager(bucket, s3_prefix=s3_prefix, snapshot_prefix=prefix)
    zfs_mgr = ZFSSnapshotManager(fs_name=filesystem, snapshot_prefix=snapshot_prefix)
    pair_manager = PairManager(s3_mgr, zfs_mgr)
    snap_name = "{}@{}".format(filesystem, snapshot)
    pair_manager.restore(snap_name, dry_run=dry)


def main():
    cfg = get_config()
    parser = argparse.ArgumentParser(
        description='list z3 snapshots',
    )
    parser.add_argument('--s3-prefix',
                        dest='s3_prefix',
                        default=cfg.get('S3_PREFIX', 'z3-backup/'),
                        help='S3 key prefix, defaults to z3-backup/')
    parser.add_argument('--filesystem', '--dataset',
                        dest='filesystem',
                        default=cfg.get('FILESYSTEM'),
                        help='the zfs dataset/filesystem to operate on')
    parser.add_argument('--snapshot-prefix',
                        dest='snapshot_prefix',
                        default=cfg.get('SNAPSHOT_PREFIX', 'zfs-auto-snap:daily'),
                        help='only operate on snapshots that start with this prefix')
    subparsers = parser.add_subparsers(help='sub-command help', dest='subcommand')

    backup_parser = subparsers.add_parser(
        'backup', help='backup local zfs snapshots to an s3 bucket')
    backup_parser.add_argument('--snapshot', dest='snapshot', default=None,
                               help='Snapshot to backup. Defaults to latest.')
    backup_parser.add_argument('--dry', dest='dry', default=False, action='store_true',
                               help='Snapshot to backup. Defaults to latest.')
    incremental_group = backup_parser.add_mutually_exclusive_group()
    incremental_group.add_argument(
        '--full', dest='full', action='store_true', help='Perform full backup')
    incremental_group.add_argument(
        '--incremental', dest='incremental', default=True, action='store_true',
        help='Perform incremental backup; this is the default')

    restore_parser = subparsers.add_parser('restore', help='not implemented')
    restore_parser.add_argument(
        'snapshot', help='Snapshot to backup. Defaults to latest.')
    restore_parser.add_argument('--dry', dest='dry', default=False, action='store_true',
                                help='Snapshot to backup. Defaults to latest.')
    subparsers.add_parser('status', help='show status of current backups')
    args = parser.parse_args()

    bucket = boto.connect_s3(
        cfg['S3_KEY_ID'], cfg['S3_SECRET']).get_bucket(cfg['BUCKET'])
    if args.subcommand == 'status':
        list_snapshots(bucket, s3_prefix=args.s3_prefix, snapshot_prefix=args.snapshot_prefix,
                       filesystem=args.filesystem)
    elif args.subcommand == 'backup':
        do_backup(bucket, s3_prefix=args.s3_prefix, snapshot_prefix=args.snapshot_prefix,
                  filesystem=args.filesystem, full=args.full, snapshot=args.snapshot,
                  dry=args.dry)
    elif args.subcommand == 'restore':
        restore(bucket, s3_prefix=args.s3_prefix, snapshot_prefix=args.snapshot_prefix,
                filesystem=args.filesystem, snapshot=args.snapshot, dry=args.dry)


if __name__ == '__main__':
    main()



from cStringIO import StringIO
from datetime import datetime
from Queue import Queue
from uuid import uuid4
import hashlib

import boto
import pytest

from z3.pput import (UploadSupervisor, UploadWorker, StreamHandler,
                     Result, WorkerCrashed, multipart_etag, parse_metadata,
                     retry, UploadException)
from z3.config import get_config


cfg = get_config()
_cached_sample_data = None


class ReadOnlyFile(object):
    def __init__(self, fd, allowed):
        self._fd = fd
        self._allowed = set(allowed)

    def __getattr__(self, name):
        if name in self._allowed:
            return getattr(self._fd, name)
        raise AssertionError("this file-like-object is readonly, {} is now allowed".format(name))


@pytest.fixture()
def sample_data():
    """Sets up a file-like-object with 6 Mbytes of data
    Since this is expensive to do, we share this object across test runs and just
    seek the file back to the start after each use.
    """
    global _cached_sample_data
    if _cached_sample_data is None:
        data = StringIO()
        chars = "".join(chr(i) for i in xrange(256))
        for count in xrange(6):
            cc = chr(count)
            for _ in xrange(2 * 1024):
                # each iteration adds 1MB
                # each 1MB chunk is made up of an alternation of the block's index (zero based)
                # and an incrementing counter (overflows to 0 several times)
                # the first block will be: 00 00 00 01 00 02 ... 00 ff 00 00 ... 00 ff
                data.write(
                    "".join(cc+chars[i] for i in xrange(256))
                )
        print "wrote {} MB" .format(data.tell() / 1024.0 / 1024.0)
        # give the test a read-only file to avoid accidentally modifying the data between tests
        _cached_sample_data = ReadOnlyFile(data, allowed=['read', 'seek'])
    _cached_sample_data.seek(0)
    return _cached_sample_data


def test_multipart_etag(sample_data):
    digests = []
    while True:
        chunk = sample_data.read(5 * 1024 * 1024)
        if len(chunk) == 0:
            break
        digest = hashlib.md5()
        digest.update(chunk)
        digests.append(digest.hexdigest())
    assert multipart_etag(digests) == '"d229c1fc0e509475afe56426c89d2724-2"'


def test_stream_handler():
    stream_handler = StreamHandler(StringIO("aabbccdde"), chunk_size=2)
    chunks = []
    while not stream_handler.finished:
        chunk = stream_handler.get_chunk()
        chunks.append(chunk)
    assert chunks == ['aa', 'bb', 'cc', 'dd', 'e']


def test_handle_results():
    sup = UploadSupervisor(None, None, None)
    sup.inbox = Queue()
    sup._pending_chunks = 3
    sup.inbox.put(Result(success=True, traceback=None, index=1, md5='a'))
    sup.inbox.put(Result(success=True, traceback=None, index=3, md5='c'))
    sup.inbox.put(Result(success=True, traceback=None, index=2, md5='b'))
    sup._handle_results()
    assert sorted(sup.results) == [(1, 'a'), (2, 'b'), (3, 'c')]
    assert sup._pending_chunks == 0


class FakeMultipart(object):
    def __init__(self, name):
        self._name = name
        self.id = str(uuid4())
        self.key_name = str(uuid4())
        self._completed = False
        self._canceled = False

    def complete_upload(self):
        if self._completed:
            raise AssertionError('multipart already completed')
        self._completed = True

    def cancel_upload(self):
        if self._canceled:
            raise AssertionError('multipart already canceled')
        self._canceled = True


class FakeBucket(object):
    def __init__(self):
        self._multipart = None

    def initiate_multipart_upload(self, name, headers):
        self._multipart = FakeMultipart(name)
        return self._multipart


class DummyWorker(UploadWorker):
    def upload_part(self, index, chunk):
        return hashlib.md5(chunk).hexdigest()


def test_supervisor_loop(sample_data):
    stream_handler = StreamHandler(sample_data)
    bucket = FakeBucket()
    sup = UploadSupervisor(stream_handler, 'test', bucket=bucket)
    etag = sup.main_loop(worker_class=DummyWorker)
    assert etag == '"d229c1fc0e509475afe56426c89d2724-2"'
    assert bucket._multipart._completed


def test_zero_data(sample_data):
    stream_handler = StreamHandler(StringIO())
    bucket = FakeBucket()
    sup = UploadSupervisor(stream_handler, 'test', bucket=bucket)
    with pytest.raises(UploadException):
        sup.main_loop(worker_class=DummyWorker)
    assert bucket._multipart._canceled is True


class ErrorWorker(UploadWorker):
    def upload_part(self, index, chunk):
        if index == 2:
            raise Exception("Testing worker crash")
        return hashlib.md5(chunk).hexdigest()


def test_supervisor_loop_with_worker_crash(sample_data):
    stream_handler = StreamHandler(sample_data)
    bucket = FakeBucket()
    sup = UploadSupervisor(stream_handler, 'test', bucket=bucket)
    with pytest.raises(WorkerCrashed):
        sup.main_loop(worker_class=ErrorWorker)


class BoomException(Exception):
    pass


class Boom(object):
    def __init__(self):
        self.count = 0

    @retry(3)
    def call(self):
        self.count += 1
        raise BoomException("Boom!")


def test_retry_decorator():
    boom = Boom()
    with pytest.raises(BoomException) as excp_info:
        for _ in xrange(3):
            boom.call()
    assert boom.count == 3


@pytest.mark.with_s3
def test_integration(sample_data):
    cfg = get_config()
    stream_handler = StreamHandler(sample_data)
    bucket = boto.connect_s3(
        cfg['S3_KEY_ID'], cfg['S3_SECRET']).get_bucket(cfg['BUCKET'])
    key_name = "z3_test_" + datetime.now().strftime("%Y%m%d_%H-%M-%S")
    sup = UploadSupervisor(
        stream_handler,
        key_name,
        bucket=bucket,
        headers=parse_metadata(["ana=are+mere", "dana=are=pere"])
    )
    etag = sup.main_loop()
    uploaded = bucket.get_key(key_name)
    assert etag == '"d229c1fc0e509475afe56426c89d2724-2"'
    assert etag == uploaded.etag
    assert uploaded.metadata == {"ana": "are+mere", "dana": "are=pere"}



from setuptools import setup, find_packages


setup(
    name="z3",
    version="0.1.6",
    platforms='any',
    packages=find_packages(),
    include_package_data=True,
    install_requires=["boto==2.28.0"],
    author="PressLabs SRL",
    author_email="contact@presslabs.com",
    url="https://github.com/presslabs/z3",
    description="Backup ZFS snapshots to S3",
    entry_points={
        'console_scripts': [
            'pput = z3.pput:main',
            'z3 = z3.snap:main',
            'z3_get = z3.get:main',
            'z3_ssh_sync = z3.ssh_sync:main'
        ]
    },
    keywords='ZFS backup',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: System :: Archiving :: Backup",
        "Topic :: Utilities",
    ],
)
